from __future__ import annotations

from typing import Callable, Annotated
from functools import wraps
import time
import logging
import threading
from pydantic import PositiveFloat, NonNegativeFloat, Field

StrictFraction: type = Annotated[float, Field(gt=0, lt=1)]

def cached_method(
        fun: Callable | None = None,
        *,
        expire_after: PositiveFloat = 4.35e+17
):

    def decorator(f):
        
        @wraps(f)
        def wrapper(obj, *args, **kwargs):
         
            key = args, tuple(kwargs.items())
 
            with wrapper.lock: 

                last_update, last_result = wrapper.cache.get(key, (0, None))
                
                if time.time() - last_update > expire_after:
                    last_result = f(obj, *args)
                    last_update = time.time()
                    wrapper.cache[key] = (last_update, last_result)
                
            return last_result

        wrapper.cache = {}
        wrapper.lock = threading.Lock()

        return wrapper

    return decorator if fun is None else decorator(f)

def sleep(
        dt: PositiveFloat,
        sleep_overhead: NonNegativeFloat = 1e-3,
        sleep_fraction: StrictFraction = 0.6,
):
    """
Semi-precise sleep function with about microsond accuracy.
 
Arguments

    dt [float, s]: 
        Time interval. 

Optional keyword arguments:

    sleep_overhead [float, s]
        Maximum overheads associated with the sleep loop.  Depends on CPU speed
        and load, YMMV.

    sleep_fraction [float]
        Once the function has finished its periodic call sleep for a 
        fraction of the remaining time to avoid saturating the processor.

Return value

    overshoot [float, s]
        Overshoot time 

    """
    t_target = time.time() + dt - call_overhead
    while (remaining := t_target - time.time()) >= 0:
        if remaining > sleep_overhead:
            time.sleep(remaining * sleep_fraction)
    return -remaining

def sleep_until(
        t: float,
        sleep_overhead: NonNegativeFloat = 1e-3,
        sleep_fraction: StrictFraction = 0.6,
):
    """
Semi-precise sleep function with about microsond accuracy.

Arguments

    t [float, s]: 
        Seconds ellapsed since epoch start, as given by time.time() 

Optional keyword arguments:

    sleep_overhead [float, s]
        Maximum overheads associated with the sleep loop.  Depends on CPU speed
        and load, YMMV.

    sleep_fraction [float]
        Once the function has finished its periodic call sleep for a 
        fraction of the remaining time to avoid saturating the processor.

Return value

    overshoot [float, s]
        Overshoot time 

    """
    while (remaining := t - time.time()) >= 0:
        if remaining > sleep_overhead:
            time.sleep(remaining * sleep_fraction)
    return -remaining 

def periodic_call(
        fun: Callable = None,
        *,
        interval: PositiveFloat = 1,
        initial_delay: NonNegativeFloat = 0.01,
        loop_condition: Callable[[object, int], bool] = lambda r, n: True,
        sleep_fraction: StrictFraction = 0.5, 
        sleep_overhead: NonNegativeFloat = 1e-3,
        call_overhead: NonNegativeFloat = 5e-6,
        threaded: bool = False
):
    """
Periodically calls a function using semi-precise timing witout saturating
processor.  

Optional keyword arguments:

    interval [float, s]
        Interval to call periodically the fonction. 

    initial_delay [float, s]
        Initial delay before the first function call

    loop_condition [function]: 
        A function that takes the result of the function (can be None) and 
        the number of calls made and returns a boolean value. By default, 
        loop forever.  

    sleep_fraction [float]
        Once the function has finished its periodic call sleep for a 
        fraction of the remaining time to avoid saturating the processor.

    sleep_overhead [float, s]
        Maximum overheads associated with the sleep loop.  Depends on CPU speed
        and load, YMMV.

    call_overhead [float, s]
        Typical overhead for a function call and return 

Example use:

    from obstechutils.precise_timing import periodic_call

    @periodic_call(
        interval=30, initial_delay=1,
        loop_condition: lambda res, n: n <= 10
    )
    def f():
        ... 

    if __name__ == "__main__":
        # operation to be repeated every 30s exactly 10 times at times 
        #    t = time.time() + 1 + 30 * n
        # with an offset and variance of microseconds or less depending 
        # on machine and CPU load.  Some outliers with milliseconds may
        # occur, though.
        f()                          

"""
    def decorator(f):

        @wraps(f)
        def wrapper(*args, **kwargs):

            fname = f.__name__
            logger = logging.getLogger('obstechutils')
            logger.info(f'{fname} periodic call every {interval:.4f} s')
          
            n = 0
            t0 = time.time() + initial_delay - call_overhead
            t_target = t0

            while True:

                delay = sleep_until(
                    t_target, 
                    sleep_overhead=sleep_overhead,
                    sleep_fraction=sleep_fraction
                )

                res = f(*args,**kwargs)
                n += 1
                logger.debug(f'{fname} call #{n} dt={delay*1e6:.3f} μs')

                if not loop_condition(res, n):
                    break
                t_target = t0 + n * interval

        if not threaded:
            return wrapper
           
        @wraps(wrapper) 
        def threaded_wrapper(*args, **kwargs):
            t = threading.Thread(
                    target=wrapper, args=args, kwargs=kwargs,
                    daemon=True
            )
            t.start()

        return threaded_wrapper

    if fun:
        return decorator(fun)
    
    return decorator
